/* Parallel Hierarchical Grid -- an adaptive finite element library.
 *
 * Copyright (C) 2005-2010 State Key Laboratory of Scientific and
 * Engineering Computing, Chinese Academy of Sciences. */

/* This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA */

/* Linear solver test with a MATLAB matrix (e.g., created by phgMatDumpMATLAB)
 * $Id: solver_test_matlab.c,v 1.12 2019/03/06 05:47:00 zlb Exp $ */

#include "phg.h"
#include <string.h>
#include <math.h>

int
main(int argc, char *argv[])
{
    VEC *x, *x0;
    MAT *A;
    SOLVER *solver;
    INT i, n, m, nnz;
    double buf[3];
    char *cmd = NULL;
    const char *magic = "solver_test_matlab: start matrix data";
    FILE *pipe = NULL;

    phgInit(&argc, &argv);

    phgPrintf("Linear solver test using a MATLAB matrix.\n");

    if (argc != 3) {
	phgPrintf("Usage:\n");
	phgPrintf("   %s [PHG_options ]Fname.m Vname[ PHG_options]\n", argv[0]);
	phgPrintf("where:\n");
	phgPrintf("  'Fname.m' is the MATLAB file defining the matrix.\n");
	phgPrintf("  'Vname' is the MATLAB variable for the matrix.\n");
	phgFinalize();
	exit(1);
    }

    if (phgRank == 0) {
	char *line = NULL;
	size_t len = 0;
	/* construct MATLAB commands for piping the matrix */
	cmd = phgAlloc(strlen(argv[1]) + strlen(argv[2]) + 1024);
	sprintf(cmd, "echo \""
		 	"source %s; [n,m]=size(%s); [r,c,v]=find(%s); "
				/*"stdout=fopen('tmp.dat', 'w'); "*/
			"fprintf(stdout, '\\n%s\\n'); "
			"fwrite(stdout, [n m length(c)], 'real*8'); "
		 	"fwrite(stdout, [r'; c'; v'], 'real*8'); "
				/*"fclose(stdout);"*/
		 "\" | /usr/bin/octave --no-gui",
		argv[1], argv[2], argv[2], magic);

	phgPrintf("Processing file \"%s\" with octave.\n", argv[1]);
	if ((pipe = popen(cmd, "r")) == NULL) {
	    fprintf(stderr, "Execution of the following command failed.\n");
	    fprintf(stderr, "%s\n", cmd);
    read_err:
	    phgAbort(1);
	    exit(1);
	}

	/* skip data before 'magic' */
	while (TRUE) {
	    if (getline(&line, &len, pipe) == -1L)
		goto read_err;
	    assert(line != NULL && line[strlen(line) - 1] == '\n');
	    line[strlen(line) - 1] = '\0';
	    if (!strcmp(line, magic))
		break;
	}
	phgFree(line);

	/* matrix size */
	assert(sizeof(double) == 8);
	if (fread(buf, sizeof(buf[0]), 3, pipe) != 3) {
	    fprintf(stderr, "Error reading matrix size.\n");
	    phgAbort(1);
	    exit(1);
	}
    }

    /* Create matrix */
#if USE_MPI
    MPI_Bcast(buf, 3, MPI_INT, 0, phgComm);
#endif	/* USE_MPI */
    n = (INT)(buf[0] + 0.5);
    m = (INT)(buf[1] + 0.5);
    nnz = (INT)(buf[2] + 0.5);
    if (n != m) {
	fprintf(stderr, "Non square matrix (%"dFMT"x%"dFMT"), abort.\n", n, m);
	phgAbort(1);
	exit(1);
    }
    m = n / phgNProcs + (phgRank < n % phgNProcs ? 1 : 0);	/* local size */
    A = phgMatCreate(phgComm, m, n);

    if (phgRank == 0 && nnz > 0) {
	/* Read matrix */
	phgPrintf("Reading matrix data.\n");
	for (i = 0; i < nnz; i++) {
	    if (fread(buf, sizeof(buf[0]), 3, pipe) != 3) {
		fprintf(stderr, "Error reading matrix data.\n");
		phgAbort(1);
		exit(1);
	    }
	    assert(buf[0] >= 0.99 && buf[1] >= 0.99);
	    if (buf[2] == 0.0)
		continue;
	    phgMatAddGlobalEntry(A, (INT)(buf[0] + 0.5) - 1,
				    (INT)(buf[1] + 0.5) - 1,
				    (FLOAT)buf[2]);
	}
	pclose(pipe); pipe = NULL;
	free(cmd); cmd = NULL;
    }

    phgMatAssemble(A);
#if USE_MPI
    buf[1] = (double)A->nnz_d + (double)A->nnz_o;
    MPI_Reduce(&buf[1], &buf[0], 1, MPI_DOUBLE, MPI_SUM, 0, phgComm);
#else	/* USE_MPI */
    buf[0] = (double)A->nnz_d + (double)A->nnz_o;
#endif	/* USE_MPI */
    phgPrintf("Matrix size: %"dFMT", nnz: %0.0lf\n", n, buf[0]);

    solver = phgMat2Solver(SOLVER_DEFAULT, A);
    phgMatDestroy(&A);

    /* setup test solution and RHS */
    x = phgMapCreateVec(solver->rhs->map, 1);
    x0 = phgMapCreateVec(solver->rhs->map, 1);
    for (i = 0; i < m; i++)
	x0->data[i] = 1.0;
    phgMatVec(MAT_OP_N, 1.0, solver->mat, x0, 0.0, &solver->rhs);
    phgPrintf("Solving linear system: ");
    phgSolverVecSolve(solver, FALSE, x);
    phgVecAXPBY(1.0, x0, -1.0, &x);
    phgPrintf("nits = %d, residual = %le, error = %le\n",
		solver->nits, solver->residual,
		(double)phgVecNorm2(x, 0, NULL));

    phgSolverDestroy(&solver);
    phgFinalize();

    return 0;
}
